
from enum import Enum, IntEnum
import hashlib
import typing

class AuthKey:

    class Type(Enum):
        Generated       : int = 0
        Temporary       : int = 1
        ReadFromFile    : int = 2
        Local           : int = 3

    def __init__(self, key : bytes, type : int = 0, dcId : int = 0) -> None:
        self.__type = type
        self.__dcId = dcId
        self.__key = key
        # if (type == self.Type.Generated) or (type == self.Type.Temporary):
            # self.__creationtime = ...
        self.__countKeyId()

    def __countKeyId(self) -> None:
        hash = hashlib.sha1(self.__key).digest()
        self.__keyId = int.from_bytes(hash[12:12+8], 'little')
    
    def prepareAES_oldmtp(self, msgKey : bytes, send : bool) -> typing.Tuple[bytes, bytes]:
        x = 0 if send else 8
        sha1_a = hashlib.sha1(msgKey[:16] + self.__key[x : x+32]).digest()
        sha1_b = hashlib.sha1(self.__key[x+32 : x+32+16] + msgKey[:16] + self.__key[x+48 : x+48+16]).digest()
        sha1_c = hashlib.sha1(self.__key[x+64 : x+64+32] + msgKey[:16]).digest()
        sha1_d = hashlib.sha1(msgKey[:16] + self.__key[x+96 : x+96+32]).digest()

        aesKey = sha1_a[:8] + sha1_b[8:8+12] + sha1_c[4:4+12]
        aesIv = sha1_a[8:8+12] + sha1_b[:8] + sha1_c[16:16+4] + sha1_d[:8]

        return aesKey, aesIv

    @property
    def key(self):
        return self.__key



from storage import *
from exception import *
from authkey import *

from PyQt5.QtCore import QByteArray, QDataStream, QBuffer, QFile, QIODevice


class TDesktop:
    def __init__(self) -> None:
        pass
    
    @property
    def passcodeKey(self) -> AuthKey:
        return self.__passcodeKey

    @property
    def localKey(self) -> AuthKey:
        return self.__localKey
        
    @property
    def account_count(self) -> int:
        return self.__account_count

    @property
    def authKey(self) -> AuthKey:
        return self.__authKey

    def FromTData(self, basePath : str, passcode : str = "", keyFile : str = "data") -> None:
        
        # READ KEY_DATA
        keyData = ReadFile("key_" + keyFile, basePath)

        salt, keyEncrypted, infoEncrypted = QByteArray(), QByteArray(), QByteArray()

        keyData.stream >> salt >> keyEncrypted >> infoEncrypted
        if keyData.stream.status() != 0:
            raise OpenTeleException(OpenTeleErrorCode.QDataStreamFailed, "Failed to stream keyData")

        
        self.__passcodeKey = CreateLocalKey(salt, QByteArray(passcode.encode("utf-8")))

        keyInnerData = DecryptLocal(keyEncrypted, self.__passcodeKey)
        self.__localKey = AuthKey(keyInnerData.stream.readRawData(256))

        info = DecryptLocal(infoEncrypted, self.__localKey)
        self.__account_count = info.stream.readInt32()

        mtp = ReadEncryptedFile(ToFilePart(ComputeDataNameKey(keyFile)), basePath, self.__localKey)

        blockId = mtp.stream.readInt32()
        if blockId != 75:
            raise OpenTeleException(OpenTeleErrorCode.FileInvalidMagic, "Not supported file version")

        serialized = QByteArray()
        mtp.stream >> serialized

        stream = QDataStream(serialized)
        stream.setVersion(QDataStream.Version.Qt_5_1)

        UserId = stream.readInt32()
        MainDcId = stream.readInt32()
        
        if ((UserId << 32) | MainDcId) == ~0:
            UserId = stream.readUInt64()
            MainDcId = stream.readInt32()

        if stream.status() != 0:
            raise OpenTeleException(OpenTeleErrorCode.QDataStreamFailed, "Could not read main fields from mtp authorization.")
        
        key_count = stream.readInt32()
        if stream.status() != 0:
            raise OpenTeleException(OpenTeleErrorCode.QDataStreamFailed, "Could not read keys count from mtp authorization.")
        
        for i in range(0, key_count):
            dcId = stream.readInt32()
            keyData = stream.readRawData(256)
            if dcId == MainDcId:
                self.__authKey = AuthKey(keyData, AuthKey.Type.ReadFromFile, dcId)
                break
            
        # while (not mtp.stream.atEnd()):
        #     blockId = mtp.stream.readInt32()
        #     if mtp.stream.status() != 0:
        #         raise OpenTeleException(OpenTeleErrorCode.QDataStreamFailed, "Failed to stream mtpData")
        #     print(blockId)
    




from tdesktop import *
from enum import Enum, IntEnum

class OpenTeleErrorCode(Enum):
    Success = ...
    FileNotFound = ...
    FileInvalidMagic = ...
    FileInvalidCheckSum = ...
    QDataStreamFailed = ...
    BadDecryptKey = ...
    BadEncryptedDataSize = ...
    BadDecryptedDataSize = ...

class OpenTeleError:


    def __init__(self, code : int) -> None:
        self._code = code
        
        error_dict = {
            OpenTeleErrorCode.Success                   : "Success without error",
            OpenTeleErrorCode.FileNotFound              : "File not found",
            OpenTeleErrorCode.FileInvalidMagic          : "File has invalid magic data",
            OpenTeleErrorCode.FileInvalidCheckSum       : "File has invalid checksum",
            OpenTeleErrorCode.QDataStreamFailed         : "QDataStream failed",
            OpenTeleErrorCode.BadDecryptKey             : "Bad decrypt key",
            OpenTeleErrorCode.BadEncryptedDataSize      : "Bad encrypted part size",
            OpenTeleErrorCode.BadDecryptedDataSize      : "Bad decrypted part size",
        }

        try:
            self._text = error_dict[code]
        except:
            raise BaseException(f"Error code is invalid: {code}")

    @property
    def text(self) -> str:
        return self._text
    
    def __str__(self) -> str:
        return self._text


class OpenTeleException(Exception):
    def __init__(self, code : OpenTeleErrorCode, *args: object) -> None:
        super().__init__(*args)
        self._error_code = OpenTeleError(code)
        self._message = args[0]

    @property
    def error_code(self) -> OpenTeleError:
        return self._error_code
    @property
    def message(self) -> OpenTeleError:
        return self._message
    
    def __str__(self):
        return f"{self._error_code.__str__()}: {self._message}"





# from opentele import TDesktop

from PyQt5.QtCore import QByteArray, QDataStream, QBuffer, QFile, QIODevice
from exception import *
import hashlib

from authkey import *
import tgcrypto

class FileReadDescriptor:
    def __init__(self) -> None:
        self.__version = 0
        self.__data = QByteArray()
        self.__buffer = QBuffer()
        self.__stream = QDataStream()

    @property
    def data(self) -> QByteArray:
        return self.__data
    @data.setter
    def data(self, value):
        self.__data = value

    @property
    def version(self) -> int:
        return self.__version
    @version.setter
    def version(self, value):
        self.__version = value

    @property
    def buffer(self) -> QBuffer:
        return self.__buffer

    @property
    def stream(self) -> QDataStream:
        return self.__stream

class EncryptedDescriptor:
    def __init__(self) -> None:
        self.__data = QByteArray()
        self.__buffer = QBuffer()
        self.__stream = QDataStream()

    @property
    def data(self) -> QByteArray:
        return self.__data
    @data.setter
    def data(self, value):
        self.__data = value

    @property
    def buffer(self) -> QBuffer:
        return self.__buffer

    @property
    def stream(self) -> QDataStream:
        return self.__stream



def ReadFile(fileName : str, basePath : str) -> FileReadDescriptor:

    to_try = ['s', '1', '0']
    tries_exception = None
    for chr in to_try:
        try:

            file = QFile(basePath + fileName + chr)
            if (not file.open(QIODevice.OpenModeFlag.ReadOnly)):
                tries_exception = OpenTeleException(OpenTeleErrorCode.FileNotFound, f"Could not open {fileName}")
                continue

            
            
            magic = file.read(4)
            if magic != b"TDF$":
                tries_exception = OpenTeleException(OpenTeleErrorCode.FileInvalidMagic, f"Invalid magic {magic} in file {fileName}")
                file.close()
                continue
            
            version = int.from_bytes(file.read(4), 'little')

            bytesdata = QByteArray(file.read(file.size()))

            dataSize = bytesdata.size() - 16

            check_md5 = bytesdata.data()[:dataSize]
            check_md5 += int(dataSize).to_bytes(4, 'little')
            check_md5 += int(version).to_bytes(4, 'little')
            check_md5 += magic
            check_md5 = hashlib.md5(check_md5)
            
            md5 = bytesdata.data()[dataSize:]
            
            if (check_md5.hexdigest() != md5.hex()):
                tries_exception = OpenTeleException(OpenTeleErrorCode.FileInvalidCheckSum, f"Invalid checksum {check_md5.hexdigest()} in file {fileName}")
                file.close()
                continue


            result = FileReadDescriptor()
            result.version = version

            bytesdata.resize(dataSize)
            result.data = bytesdata
            bytesdata = QByteArray()

            result.buffer.setBuffer(result.data)

            result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
            result.stream.setDevice(result.buffer)
            result.stream.setVersion(QDataStream.Version.Qt_5_1)

            file.close()
            return result
        except IOError as e:
            pass
    
    raise tries_exception if tries_exception else OpenTeleException(OpenTeleErrorCode.FileNotFound, f"Could not open {fileName}")

def ReadEncryptedFile(fileName : str, basePath : str, authKey : AuthKey) -> FileReadDescriptor:

    result = ReadFile(fileName, basePath)
    encrypted = QByteArray()
    result.stream >> encrypted

    try:
        data = DecryptLocal(encrypted, authKey)
    except OpenTeleException as e:
        result.stream.setDevice(None)
        if result.buffer.isOpen(): result.buffer.close()
        result.buffer.setBuffer(None)
        result.data = QByteArray()
        result.version = 0
        raise e
    
    result.stream.setDevice(None)
    if result.buffer.isOpen(): result.buffer.close()
    result.buffer.setBuffer(None)
    result.data = data.data
    
    result.buffer.setBuffer(result.data)
    result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
    result.buffer.seek(data.buffer.pos())
    result.stream.setDevice(result.buffer)
    result.stream.setVersion(QDataStream.Version.Qt_5_1)

    return result

    


def CreateLocalKey(salt : QByteArray, passcode: QByteArray = QByteArray()) -> AuthKey:
    hashKey = hashlib.sha512(salt)
    hashKey.update(passcode)
    hashKey.update(salt)

    iterationsCount = 1 if passcode.isEmpty() else 100000
    return AuthKey(hashlib.pbkdf2_hmac("sha512", hashKey.digest(), salt, iterationsCount, 256))


def aesDecryptLocal(src : QByteArray, authKey : AuthKey, key128 : QByteArray) -> QByteArray:
    aesKey, aesIv = authKey.prepareAES_oldmtp(key128, False)
    decrypted = tgcrypto.ige256_decrypt(src, aesKey, aesIv)
    return QByteArray(decrypted)

def DecryptLocal(encrypted : QByteArray, authKey : AuthKey) -> EncryptedDescriptor:
    encryptedSize = encrypted.size()
    if (encryptedSize <= 16) or (encryptedSize & 0x0F):
        print(encrypted.data().hex())
        raise OpenTeleException(OpenTeleErrorCode.BadEncryptedDataSize, f"Bad encrypted part size: {encryptedSize}")

    fullLen = encryptedSize - 16
    encryptedKey = encrypted[:16]
    encryptedData = encrypted[16:]

    decrypted = aesDecryptLocal(encryptedData, authKey, encryptedKey)
    checkHash = hashlib.sha1(decrypted).digest()[:16]
    if (checkHash != encryptedKey):
        raise OpenTeleException(OpenTeleErrorCode.BadDecryptKey, "Bad decrypt key, data not decrypted - incorrect password?")

    dataLen = int.from_bytes(decrypted[:4], 'little') # *(const uint32*)decrypted.constData();
    if (dataLen > decrypted.size()) or (dataLen <= fullLen - 16) or (dataLen < 4):
        raise OpenTeleException(OpenTeleErrorCode.BadDecryptedDataSize, f"Bad decrypted part size: {encryptedSize}, fullLen: {fullLen}, decrypted size: {decrypted.__len__()}")

    decrypted.resize(dataLen)
    result = EncryptedDescriptor()
    result.data = decrypted

    decrypted = QByteArray()
    
    result.buffer.setBuffer(result.data)
    result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
    result.buffer.seek(4) # skip len
    result.stream.setDevice(result.buffer)
    result.stream.setVersion(QDataStream.Version.Qt_5_1)

    return result


def ComputeDataNameKey(dataName : str) -> int:
    md5 = hashlib.md5(dataName.encode("utf-8")).digest()
    return int.from_bytes(md5, 'little')

def ToFilePart(val : int):
    result = str()
    for i in range(0, 0x10):
        v = val & 0xF
        if v < 0x0A:
            result += chr(ord('0') + v)
        else:
            result += chr(ord('A') + (v - 0x0A))
        val >>= 4
    return result